require 'asn1-pure.rb'
require 'yaml'
require 'time'
require 'cbor-diagnostic'

class String
  def hexi
    bytes.map{|x| "%02x" % x}.join
  end
  def hexs
    bytes.map{|x| "%02x" % x}.join(" ")
  end
  def xeh
    gsub(/\s/, "").chars.each_slice(2).map{ |x| Integer(x.join, 16).chr("BINARY") }.join
  end
end

if ARGV[0] == "-b"
  ARGV.shift
  $binary = true
end
DUMPFILE = "dumped-cert.yml"
if ARGV[0] == "-d"
  ARGV.shift
  $dump = true
end
if ARGV[0] == "-b"
  ARGV.shift
  $binary = true
end
f = ARGF.read.b

def numbertobytes(n)
  if n == 0
    "".b
  else
    s = n.to_s(16)
    s[0...0] = "0" if s.size.odd?
    [s].pack("H*").b
  end
end
fail unless numbertobytes(0) == ""
fail unless numbertobytes(10) == "\x0A"
fail unless numbertobytes(160) == "\xA0".b
fail unless numbertobytes(1000) == "\x03\xE8".b

def bitstringtobytes(s)
  fail "bitstring #{s}" if s[0] != "\0"
  s[1..-1]
end

def datetimeasn1(s)
  case s.keys
  when [:t]
    Time.iso8601(s.fetch(:t).sub(/\A(..)(..)(..)(..)(..)(..)Z\z/){"20#$1-#$2-#$3T#$4:#$5:#$6Z"}).to_i
  when [:gt]
    Time.iso8601(s.fetch(:gt).sub(/\A(....)(..)(..)(..)(..)(..(?:[.]\d*)?)Z\z/){"#$1-#$2-#$3T#$4:#$5:#$6Z"}).to_i
  else
    fail ["unknown date/time type", s.keys].inspect
  end
end

OIDLOOKUP = Hash.new {|h, k| k}
%w(
2.5.4.3                              CN
1.2.840.113549.1.9.1                 E
2.5.4.11                             OU
2.5.4.10                             O
2.5.4.7                              L
2.5.4.8                              ST
2.5.4.6                              C
0.9.2342.19200300.100.1.25           DC

2.5.4.12                             TITLE
0.9.2342.19200300.100.1.3            MAIL
2.5.4.5                              SERIALNUMBER
1.2.840.113549.1.9.2                 UNSTRUCTUREDNAME
1.2.840.113549.1.9.8                 UNSTRUCTUREDADDRESS
2.5.4.42                             GN
2.5.4.4                              SN

2.5.4.9                              STREET
0.9.2342.19200300.100.1.1            UID
).each_slice(2).map{ |x, y| OIDLOOKUP[x] = y}
# middle group stolen from netscape certificate management system administrator guide
# puts OIDLOOKUP.to_yaml

def dntostring(dn)              # almost RFC 4514
  sets = dn.fetch(:seq)
  sets.map do |s|
    g = s.fetch(:set)
    g.map do |elt|
      aname, aval = elt.fetch(:seq)
      [OIDLOOKUP[aname], aval.values.first.gsub(/[ #"+,;<>\\]/){"\\#$&"}].join("=")
    end.join("+")
  end.join(",")
end

ALGLOOKUP = Hash.new {|h, k| k}
%w(
1.2.840.113549.1.1.11 sha256WithRSAEncryption
1.2.840.113549.1.1.1  rsaEncryption
1.2.840.10045.2.1     ecPublicKey

1.2.840.10045.3.1.7   secp256r1
1.2.840.10045.3.1.34  secp384r1
1.2.840.10045.3.1.35  secp521r1
).each_slice(2).map{ |x, y| ALGLOOKUP[x] = y.intern}
# prime192v1(1)   prime192v2(2)   prime192v3(3)  prime239v1(4)   prime239v2(5)   prime239v3(6)   prime256v1(7) 

SYMLOOKUP = Hash.new {|h, k| warn "label unknown for key :#{k}"; k}
%w(
iss   1
sub   2
aud   3
exp   4
nbf   5
iat   6
cti   7


kty 1
n -1
e -2

crv -1
x -2
y -3

pk -100001
).each_slice(2).map{ |x, y| SYMLOOKUP[x.intern] = y.to_i}

StopDS = Struct.new(:value)
Bits = Struct.new(:bits)

def desymbolicate(o)
  case o
  when Symbol
    SYMLOOKUP[o]
  when Hash
    Hash[o.map {|k, v| [desymbolicate(k), desymbolicate(v)]}]
  when Array
    o.map {|x| desymbolicate(x)}
  when StopDS
    o.value
  else
    o
  end
end

#  RFC 4514

cert = ASN1.decode(f)
claimset = {}
if $dump
  warn "Dumping cert as #{DUMPFILE}..."
  File.write(DUMPFILE, cert.to_yaml)
end
tbs, sigalg, sigval = cert.fetch(:seq)
ver, ser, sigalg1, iss, validity, sub, spki, *rest = tbs.fetch(:seq)
fail [:ver, ver].inspect unless ver == {exp0: [2]}
fail [:ser, set].inspect unless Integer === ser
claimset[:cti] = numbertobytes(ser)
fail [:alg, sigalg, sigalg1].inspect unless sigalg == sigalg1

decoded_sigalg = ALGLOOKUP[sigalg.fetch(:seq).first]
# p decoded_sigalg

notbefore, notafter = validity.fetch(:seq).map{|x| datetimeasn1(x)}
# p [notbefore, notafter]
claimset[:nbf] = notbefore
claimset[:exp] = notafter

claimset[:iss] = dntostring(iss)
claimset[:sub] = dntostring(sub)

pkalg, pkbits = spki.fetch(:seq)
pkalgseq1, pkalgseq2 = pkalg.fetch(:seq)
decoded_pkalg = ALGLOOKUP[pkalgseq1]
decoded_pkalgparm = ALGLOOKUP[pkalgseq2]
kk = bitstringtobytes(pkbits.fetch(:bits))
case decoded_pkalg
when  :rsaEncryption
  kk = ASN1.decode(kk)
  n, e = kk.fetch(:seq)
  u = CBOR.encode(n)
  claimset[:pk] = {kty: 3, n: numbertobytes(n), e: numbertobytes(e)}
when :ecPublicKey
  case decoded_pkalgparm
  when :secp256r1
    case kk.getbyte(0)
    # XXX: check lengths
    when 4
      fail "Weird EC key #{kk.hexi}" unless kk.size == 65
      y = kk[33..64]
    when 2
      fail "Weird EC key #{kk.hexi}" unless kk.size == 33
      y = false
    when 3
      fail "Weird EC key #{kk.hexi}" unless kk.size == 33
      y = true
    else
      fail "Weird EC key #{kk.hexi}"
    end
    claimset[:pk] = {kty: 2, crv: 1, x: kk[1..32], y: y}
  else
    claimset[:pk] = ["ecPublicKey", pkalgseq2, kk] # TODO convert to COSE key
  end
else
  claimset[:pk] = [decoded_pkalg, pkalgseq2, kk] # TODO convert to COSE key
end

rest1 = rest.reduce({}, :merge)

exp3 = rest1.delete(:exp3)
puts [:rest1, rest1].to_yaml unless rest1 == {}

exp3[0].fetch(:seq).each do |e3|
  name, crit, val = e3.fetch(:seq)
  if Hash === crit
    val = crit
    crit = false
  end
  val = val.fetch(:b)
  val = ASN1.decode(val)
  # warn "@@ #{val.inspect}"
  if Hash === val && val.keys == [:bits]
    s = val[:bits]
    ignore_bits = s.getbyte(0)
    bits = s[1..-1].bytes.map {|b| (b+0x100).to_s(2)[1..-1] }.join
    bits[-1-ignore_bits+1..-1] = '' if ignore_bits > 0
    val = {bits: bits}
  end
  val = StopDS.new(val)
  claimset[name] = val          # er, ignoring crit right now
end

claimset = desymbolicate(claimset)

claimset_cbor = claimset.to_cbor
warn "CBOR size: #{claimset_cbor.size}"
if $binary
  print claimset_cbor
else
  puts CBOR.decode(claimset_cbor).cbor_diagnostic
end
